library webview;

import 'dart:async';
import 'dart:html';
import 'dart:json' as json;
import 'package:js/js.dart' as js;

/// A custom element to securely embed content (e.g. a web page) in a Chrome
/// application.
class Webview { // TODO(rms): extend `Element` once Custom Elements land.
   
//  static const EventStreamProvider<CustomEvent> exitEvent = 
//      const EventStreamProvider<CustomEvent>('exit');
//     
//  static const EventStreamProvider<CustomEvent> loadAbortEvent = 
//      const EventStreamProvider<CustomEvent>('loadabort');
//    
//  static const EventStreamProvider<CustomEvent> loadCommitEvent = 
//      const EventStreamProvider<CustomEvent>('loadcommit');
//    
//  static const EventStreamProvider<CustomEvent> loadRedirectEvent = 
//      const EventStreamProvider<CustomEvent>('loadredirect');
//    
//  static const EventStreamProvider<CustomEvent> loadStartEvent = 
//      const EventStreamProvider<CustomEvent>('loadstart');
//    
//  static const EventStreamProvider<CustomEvent> loadStopEvent = 
//      const EventStreamProvider<CustomEvent>('loadstop');
  
  static const Map _EVENT_DETAIL = const { 
    'exit' : const ['processId', 'reason'],
    'loadabort' : const ['url', 'isTopLevel', 'reason'],
    'loadcommit' : const ['url', 'isTopLevel'],
    'loadredirect' : const ['oldUrl', 'newUrl', 'isTopLevel'],
    'loadstart' : const ['url', 'isTopLevel'],
    'loadstop' : const [],
    'sizechanged': const ['oldHeight', 'oldWidth', 'newHeight', 'newWidth']
  };
  
  static bool _isSupported = true;
  /// Whether or not the webview element is supported by the platform.
  static bool get isSupported => _isSupported;
  
  static bool _isInjected = false;
  static _inject(then) {
    if(_isInjected) then();
    else {
      // Test if the js script is already injected.
      try {
        _isSupported = js.scoped(() => js.context.isWebviewSupported());
        // If no error is thrown above then webview.js is injected.
        _isInjected = true;
      } catch(e) {}
      
      if(_isInjected) then();
      else {
        // Webview.js is not yet injected, so append it to body.
        var script = new ScriptElement();
        script.type = "text/javascript";
        script.src = "packages/webview/webview.js";
        script.onLoad.listen((e) {
          // Script loaded, check support and continue.
          _isSupported = js.scoped(() => js.context.isWebviewSupported());
          then();
        });
        document.body.append(script);
        _isInjected = true;
      }
    }
  }
  
  String _src = '';
  /// The url of the embedded content.
  String get src => isAvailable ? _call(() => _webview.getSrc()) : _src;
         set src(value) => 
             isAvailable ? _call(() => _webview.setSrc(value)) : _src = value;
         
  /// Whether this webview instance is available.  The webview is available
  /// once it is inserted in the DOM _and_ the javascript behavior is loaded.
  bool get isAvailable => _webview != null;

  /// Whether or not this webview can navigate [back].
  bool get canGoBack => _call(() => _webview.canGoBack());
  
  /// Whether or not this webview can navigate [forward].
  bool get canGoForward => _call(() => _webview.canGoForward());
  
  _Window _contentWindow;
  /// The window object of this webview.
  WindowBase get contentWindow {
    if(_contentWindow == null) {
      _contentWindow = _call(() => new _Window(_webview.contentWindow()));
    }
    return _contentWindow;
  }

  final _onExit = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when the process rendering the webview content has exited.
  ///
  /// Details:
  ///   `int processId`:        The [processId] of the webview that exited.
  ///   `String reason`:        One of `normal`, `abnormal`, `crash`, `kill`.
  Stream<CustomEvent> get onExit => _onExit.stream;//exitEvent.forTarget(this);
  
  final _onLoadAbort = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a top-level load has aborted without committing.
  ///
  /// Details:
  ///   `String url`:           The requested url.
  ///   `boolean isTopLevel`:   Whether the load was top-level or in a subframe.
  ///   `String reason`:        One of `networkError`, `download`, `canceled`,
  ///                           `sslError`, or `safeBrowsingError`.
  Stream<CustomEvent> get onLoadAbort => _onLoadAbort.stream;//loadAbortEvent.forTarget(this);
  
  final _onLoadCommit = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a load has committed.
  ///
  /// Details:
  ///   `String url`:           The url that committed.
  ///   `boolean isTopLevel`:   Whether the load is top-level or in a subframe.
  Stream<CustomEvent> get onLoadCommit => _onLoadCommit.stream;//loadCommitEvent.forTarget(this);
  
  final _onLoadRedirect = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a top-level load request has redirected to another url.
  ///
  /// Details:
  ///   `String oldUrl`:        The requested url, before the redirect.
  ///   `String newUrl`:        The new requested url, after the redirect.
  ///   `boolean isTopLevel`:   Whether the redirect is top-level or in a 
  ///                           subframe.
  Stream<CustomEvent> get onLoadRedirect => _onLoadRedirect.stream;//loadRedirectEvent.forTarget(this);
  
  final _onLoadStart = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a load has begun.
  ///
  /// Details:
  ///   `String url`:           The requested url.
  ///   `boolean isTopLevel`:   Whether the load is top-level or in a subframe.
  Stream<CustomEvent> get onLoadStart => _onLoadStart.stream;//loadStartEvent.forTarget(this);
  
  final _onLoadStop = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when all loads in the element (including all subframes) have 
  /// completed. This event is dispatched in addition to any [loadCommitEvent] 
  /// or [loadAbortEvent], which occur per-frame.
  Stream<CustomEvent> get onLoadStop => _onLoadStop.stream;//loadStopEvent.forTarget(this);
  
  /// The process id of this webview.
  int get processId => _call(() => _webview.getProcessId());

  final Element element;
  
  js.Callback _onEvent;
  js.Proxy _webview;

  Webview() : this.wrap(new Element.tag('webview'));
  
  Webview.wrap(this.element) {
    _inject(() {
      js.scoped(() {
        _onEvent = new js.Callback.many(_dispatch);
        _webview = js.retain(
            new js.Proxy(js.context.Webview, element, _src, _onEvent));
      });
    });
  }
  
  void dispose() {
    js.scoped(() {
      _onEvent.dispose();
      js.release(_webview);
      if(_contentWindow != null) _contentWindow._dispose();
    });
  }
  
  /// Navigate back one step in this webview's history.
  void back() => _call(() => _webview.back());  
  
  /// Inject and execute a script in the webview's context.
  /// 
  /// NOTE: this is an experimental and undocumented feature.  It only works at
  /// present with javascript, and you should specify one of [code] or [file],
  /// but _not_ both.
  void executeScript({String code, String file}) {
    final m = {};
    if (code != null) m['code'] = code;
    if (file != null) m['file'] = file;
    _call(() => _webview.executeScript(js.map(m)));
  }
  
  /// Navigate forward one step in this webview's history.
  void forward() => _call(() => _webview.forward());
  
  /// Navigate to a [relativeIndex] in this webview's history.
  ///
  /// If the [relativeIndex] is positive it means to go forward that number of
  /// history entries; if negative it means to go back that number of entries.
  void go(int relativeIndex) => _call(() => _webview.go(relativeIndex));
  
  /// Reload the content of this webview.
  void reload() => _call(() => _webview.reload());  
  
  /// Stops this webview if it is currently loading content.
  void stop() => _call(() => _webview.stop());  
  
  /// Terminates the content's renderer process.
  ///
  /// This may affect multiple webviews in the current app if they share 
  /// [processId], but it will not affect webviews in other apps.
  void terminate() => _call(() => _webview.terminate());
  
  _call(f()) {
    if(!isSupported) throw new UnsupportedError('Webview is not supported.');
    if(!isAvailable) throw new StateError('Webview instance is not available.');
    return js.scoped(f);
  }
  
  bool _dispatch(e) {
    // TODO(rms): explore better ways to do this
    var detail = new Map();
    for(var d in _EVENT_DETAIL[e.type]) {      
      switch(d) {
        case 'isTopLevel' : detail['isTopLevel'] = e.isTopLevel; break;  
        case 'oldHeight' : detail['oldHeight'] = e.oldHeight; break;
        case 'oldUrl' : detail['oldUrl'] = e.oldUrl; break;
        case 'oldWidth' : detail['oldWidth'] = e.oldWidth; break;
        case 'newHeight' : detail['newHeight'] = e.newHeight; break;
        case 'newUrl' : detail['newUrl'] = e.newUrl; break;
        case 'newWidth' : detail['newWidth'] = e.newWidth; break;
        case 'processId' : detail['processId'] = e.processId; break;
        case 'reason' : detail['reason'] = e.reason; break;
        case 'url': detail['url'] = e.url; break;
      }
    }
    
    final ev = new CustomEvent(e.type, 
                               canBubble: e.bubbles, 
                               cancelable: e.cancelable, 
                               detail: json.stringify(detail));
    
    // TODO(rms): for now this is not an `Element` so we don't dispatch via DOM. 
//    e.stopPropagation();    
//    dispatchEvent(ev); 
    
    switch(ev.type) {
      case 'exit': _onExit.add(ev); break;
      case 'loadabort': _onLoadAbort.add(ev); break;
      case 'loadcommit': _onLoadCommit.add(ev); break;
      case 'loadredirect': _onLoadRedirect.add(ev); break;
      case 'loadstart': _onLoadStart.add(ev); break;
      case 'loadstop': _onLoadStop.add(ev); break;
      default: break;
    }
  }
}

class _Window implements WindowBase {  
  bool get closed => _proxy.closed;    
  HistoryBase get history => throw new UnimplementedError();  
  LocationBase get location => throw new UnimplementedError();  
  WindowBase get opener => throw new UnimplementedError();  
  WindowBase get parent => throw new UnimplementedError();  
  WindowBase get top => throw new UnimplementedError();  
  js.Proxy _proxy;  
  _Window(js.Proxy proxy) : _proxy = js.retain(proxy);  
  void dispose() => js.release(_proxy);  
  void close() => _proxy.close(); 
  void postMessage(message, String targetOrigin, [List messagePorts]) {
      if(messagePorts != null)
        throw new UnsupportedError("messagePorts are not supported.");
      _proxy.postMessage(message, targetOrigin);
  }
}
